#include <sysexits.h>
#include <stdlib.h>
#include <string.h>

#ifdef __OS2__
#define BSD_SELECT
#include <types.h>
#include <netinet\in.h>
#include <sys\socket.h>
#include <sys\select.h>
#define SOCKET int
#define bzero(p,b) memset(p,0,b)
#include <netdb.h>
#endif
#ifdef AIX
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/select.h>
#define SOCKET int
#include <netdb.h>
#define soclose(fd) close(fd)
#define psock_errno(s) perror(s)
#include <errno.h>
#endif

#ifndef NDEBUG
#define DPRINTF(d) printf##d
#else
#define DPRINTF(d)
#endif

#define MAXBUF 4096
void relay(SOCKET s_in, SOCKET s_out) {
        char in_buffer[MAXBUF],out_buffer[MAXBUF];
        int in_bytes=0,out_bytes=0;
        /*relay loop*/
        while(s_in>=0 || s_out>=0) {
                /*wait for something to happen*/
                int x;
                struct fd_set fds_read,fds_write;
                int nfds=0;
DPRINTF(("relay_loop: s_in=%d s_out=%d inbytes=%d outbytes=%d\n",s_in,s_out,in_bytes,out_bytes));
                
                FD_ZERO(&fds_read);
                FD_ZERO(&fds_write);
                if(s_in>=0) {
                        if(in_bytes>0)
                                FD_SET(s_in,&fds_write);
                        if(out_bytes==0)
                                FD_SET(s_in,&fds_read);
                        if(s_in>nfds) nfds=s_in;
                }        
                if(s_out>=0) {
                        if(out_bytes>0)
                                FD_SET(s_out,&fds_write);
                        if(in_bytes==0)
                                FD_SET(s_out,&fds_read);
                        if(s_out>nfds) nfds=s_out;
                }
                x = select(nfds+1, &fds_read,&fds_write, 0, 0);
                if(x<0) exit(errno);
                
                /*now read and write as necessary*/
                if(s_in>=0) {
                        if(in_bytes>0 && FD_ISSET(s_in,&fds_write)) {
                                /*write ingoing*/
                                int b;
DPRINTF(("write s_in\n"));
                                b = send(s_in, in_buffer, in_bytes, 0);
DPRINTF(("b=%d\n",b));
                                if(b<0) {
                                        soclose(s_in);
                                        s_in = -1;
                                } else {
                                        memmove(in_buffer, in_buffer+b, in_bytes-b);
                                        in_bytes -= b;
                                }
                        }
                        if(out_bytes==0 && FD_ISSET(s_in,&fds_read)) {
                                /*read ingoing*/
                                int b;
DPRINTF(("read s_in\n"));
                                b = recv(s_in, out_buffer, MAXBUF, 0);
DPRINTF(("b=%d\n",b));
                                if(b<=0) {
                                        soclose(s_in);
                                        s_in = -1;
                                } else
                                        out_bytes = b;
                        }
                }
                if(s_out>=0) {
                        if(out_bytes>0 && FD_ISSET(s_out,&fds_write)) {
                                /*write outgoing*/
                                int b;
DPRINTF(("write s_out\n"));
                                b = send(s_out, out_buffer, out_bytes, 0);
DPRINTF(("b=%d\n",b));
                                if(b<0) {
                                        soclose(s_out);
                                        s_out = -1;
                                } else {
                                        memmove(out_buffer, out_buffer+b, out_bytes-b);
                                        out_bytes -= b;
                                }
                        }
                        if(in_bytes==0 && FD_ISSET(s_out,&fds_read)) {
                                /*read ingoing*/
                                int b;
DPRINTF(("read s_out\n"));
                                b = recv(s_out, in_buffer, MAXBUF, 0);
DPRINTF(("b=%d\n",b));
                                if(b<=0) {
                                        soclose(s_out);
                                        s_out = -1;
                                } else {
                                        in_bytes = b;
                                }
                        }
                }
                if(s_out<0) out_bytes=0;
                if(s_in<0) in_bytes=0;
                if(s_in<0 && out_bytes==0) {
                        soclose(s_out);
                        s_out=-1;
                }        
                if(s_out<0 && in_bytes==0) {
                        soclose(s_in);
                        s_in=-1;
                }
                
        }
}

unsigned short str2port(const char *s) {
        /* Convert ascii port/service name to port number*/
        char *endptr;
        unsigned short port=(unsigned short)strtol(s,&endptr,0);
        if(endptr && *endptr) {
                /* not digits */
                struct servent *se=getservbyname((char*)s,"tcp");
                if(se)
                        return se->s_port;
                else
                        return 0;
        } else
                return htons(port);
}

unsigned long str2inetaddr(const char *s) {
        /* Convert ascii hostname/ipaddress to 32-bit address
         * if the string only contains digits and dots we assume it is a
         * direct ip address
         */
        const char *p;
        for(p=s; *p; p++){
                if(!((*p>='0' && *p<'9') || *p=='.')) {
                        struct hostent *he=gethostbyname((char*)s);
                        if(he)
                                return ((struct in_addr*)(he->h_addr_list[0]))->s_addr;
                        else
                                return 0;
                }
        }
        return inet_addr((char*)s);
}

int main(int argc, char **argv) {
        int s_in,s_out;
        if((argc!=3 && argc!=5)
           || strcmp(argv[1],"-usage")==0
           || strcmp(argv[1],"+help")==0)
        {
                printf("tcprelay - TCP/IP relay\n"
                       "usage: tcprelay forwardhost forwardport [listenhost listenport]\n"
                      );
                return 1;      
        }
        
        if(argc==5
#ifdef __OS2__
           && strcmp(argv[3],"!")!=0
#endif
          )
        {
                /*listen for client to connect*/
                int s;
                struct sockaddr_in myaddress;
                struct sockaddr client;
                int clientnamelen;
                
                s = socket(PF_INET, SOCK_STREAM, 0);
                if(s<0) {
                        psock_errno(argv[0]);
                        return EX_OSERR;
                }
                
                memset(&myaddress,0,sizeof(myaddress));
                myaddress.sin_family = AF_INET;
                if(strcmp(argv[3],"-")!=0)
                        myaddress.sin_addr.s_addr = str2inetaddr(argv[3]);
                myaddress.sin_port = str2port(argv[4]);
                if(bind(s,(struct sockaddr*)&myaddress,sizeof(myaddress))) {
                        psock_errno(argv[0]);
                        return EX_DATAERR;
                }
                if(listen(s,1)) {
                        psock_errno(argv[0]);
                        soclose(s);
                        return EX_UNAVAILABLE;
                }
                clientnamelen=sizeof(client);
                if((s_in=accept(s,&client,&clientnamelen))==-1) {
                        psock_errno(argv[0]);
                        soclose(s);
                        return EX_UNAVAILABLE;
                }
                soclose(s);
        } else {
#ifdef AIX
                s_in = 0;
#endif
#ifdef __OS2__
                s_in = atoi(argv[4]); //socket handle is argument
#endif
        }
        {
                /*connect to forward address*/
                struct sockaddr_in server;
                s_out = socket(PF_INET, SOCK_STREAM, 0);
                if(s_out<0) {
                        psock_errno(argv[0]);
                        return EX_OSERR;
                }
                memset(&server,0,sizeof(server));
                server.sin_family = AF_INET;
                server.sin_addr.s_addr = str2inetaddr(argv[1]);
                server.sin_port = str2port(argv[2]);
                if(connect(s_out,(struct sockaddr*)&server,sizeof(server))<0) {
                        psock_errno(argv[0]);
                        return EX_UNAVAILABLE;
                }
        }
        
        {
                /*setup linger*/
                struct linger lstruct;
                lstruct.l_onoff = 1;
                lstruct.l_linger = 100;
                setsockopt(s_in,  SOL_SOCKET, SO_LINGER, (char*)&lstruct, sizeof(lstruct));
                setsockopt(s_out, SOL_SOCKET, SO_LINGER, (char*)&lstruct, sizeof(lstruct));
        }        
        relay(s_in,s_out);

        return 0;
}
